
/*
   z273
   Read and write the data of a Midland Z-273 module.
 */

#include "data.h"

// Pin mappings from the Z-273 to the Arduino.  First 6 bits are
// written into the flip-flop and then the pins are reused for
// 5 more bits of the address.  These bits are first:
int AD2 = A5;
int AD3 = A4;
int AD4 = A3;
int AD5 = A2;
int AD6 = A1;
int AD7 = A0;
// And these bits come second:
int AD0 = A5;
int AD1 = A4;
int AD8 = A3;
int AD9 = A2;
int AD10 = A1;

// Pin mappings from the Z-273 to the Arduino.  These are the 8 data lines:
int DO0 = 10;
int DO1 = 9;
int DO2 = 2;
int DO3 = 3;
int DO4 = 4;
int DO5 = 5;
int DO6 = 7;
int DO7 = 8;

// Pin mapping from the Z-273 to the Arduino.  Three control lines:
int CE = 11;
int _OE = 12;	// always use helpers to manipulate OE
int CLK = 13;

// We use one pin to sense when programming voltage has been applied to VPP and
// try not to let the OE pin fall when that happens (because you can fry everything).
int VPP = 6;

// ***** program setup here *****
void setup() {
  setupAllZ273pins();

  Serial.begin(57600);
  clearSerialInputBuffer();
}

// ***** main program here *****
void loop() {
  unsigned char key;
  
  do {
    delay( 100 );
    key = Serial.read();

    if( key == 10 || key == 13 ) {
	clearSerialInputBuffer();
    }

    // Wait for user to press ? before sending data.
    if( key == '?' ) {
      if( digitalRead(VPP) == LOW) {
	readAllDataFromROM();
	clearSerialInputBuffer();
      } else {
	Serial.println( "Can't read data with programming voltage applied." );
      }
    }

    // If they hit = then we'll do a verify of the data.
    if( key == '=' ) {
      if( digitalRead(VPP) == LOW) {
	verifyAllDataFromROM();
	clearSerialInputBuffer();
      } else {
	Serial.println( "Can't verify data with programming voltage applied." );
      }
    }

    // If they hit ! then we'll write data to the chip.
    if( key == '!' ) {
      programAllDataToROM( 1 );
      clearSerialInputBuffer();
    }

    // If they hit @ then we'll write data to the chip -- 2 tries max.
    if( key == '@' ) {
      programAllDataToROM( 2 );
      clearSerialInputBuffer();
    }

    // If they hit # then we'll write data to the chip -- 2 tries max.
    if( key == '#' ) {
      programAllDataToROM( 3 );
      clearSerialInputBuffer();
    }
  } while( true );
}


void clearSerialInputBuffer() {
  while( Serial.available() > 0 ) {
    Serial.read();
  }

  Serial.println("Press '?' for data, '=' to verify, and '!' to burn:");
  if( digitalRead(VPP)==HIGH )
    Serial.println( "Progamming voltage is in effect.");
}


void setupAllZ273pins() {
  setupControlPinsForOutputAndDisable();
  setupAddressPinsForOutput();
  setDataPinsForInput();
}

void setOEpinHIGH() {
  digitalWrite(_OE, HIGH);
}

// Setting OE low is dangerous if the programming voltage has been applied.
void setOEpinLOW() {
  if( digitalRead(VPP) == HIGH ) {
    Serial.println( "OE will not be set LOW with programming voltage in effect." );
  } else {
    digitalWrite(_OE,LOW);
  }
}

void setupControlPinsForOutputAndDisable() {
  // Sets each Arduino pin as an output pin and puts the EPROM in output disable
  // mode or program inhibit mode (OH high, CE low).
  pinMode(_OE, OUTPUT);
  setOEpinHIGH();

  pinMode(CE, OUTPUT);
  digitalWrite(CE, LOW);

  pinMode(CLK, OUTPUT);
  digitalWrite(CLK, LOW);
}

void setupAddressPinsForOutput() {
  // Addresses are sent on the Analog pins, which we use in digital mode in this program.
  pinMode(A0, OUTPUT);
  pinMode(A1, OUTPUT);
  pinMode(A2, OUTPUT);
  pinMode(A3, OUTPUT);
  pinMode(A4, OUTPUT);
  pinMode(A5, OUTPUT);
  digitalWrite(A0, LOW);
  digitalWrite(A1, LOW);
  digitalWrite(A2, LOW);
  digitalWrite(A3, LOW);
  digitalWrite(A4, LOW);
  digitalWrite(A5, LOW);
}

void setDataPinsForOutput() {
  pinMode(DO0, OUTPUT);
  pinMode(DO1, OUTPUT);
  pinMode(DO2, OUTPUT);
  pinMode(DO3, OUTPUT);
  pinMode(DO4, OUTPUT);
  pinMode(DO5, OUTPUT);
  pinMode(DO6, OUTPUT);
  pinMode(DO7, OUTPUT);
}

void setDataPinsForInput() {
  pinMode(DO0, INPUT);
  pinMode(DO1, INPUT);
  pinMode(DO2, INPUT);
  pinMode(DO3, INPUT);
  pinMode(DO4, INPUT);
  pinMode(DO5, INPUT);
  pinMode(DO6, INPUT);
  pinMode(DO7, INPUT);
}

void readAllDataFromROM() {
  // Warning.  This routine should not be called when Vpp=25V programming voltage.
  int i, j;
  int address;
  unsigned char databyte;
  boolean chipIsErased = true;

  Serial.println();
  Serial.println( "Data follows:" );

  // Read data from the EPROM and send it to the computer in HEX.
  // We'll do this 32 bytes=64 chars per line for 64 lines.
  for( i = 0; i < 64; i++ ) {
    for( j = 0; j < 32; j++ ) {
      address = 32*i + j;
      setAddressForROMTransfer(address);

      databyte = readByteFromROM();
      printPaddedHex(databyte);

      if( databyte != 0xFF ) chipIsErased = false;
    }

    Serial.println();
  }

  if( chipIsErased ) {
    Serial.println();
    Serial.println( "This EPROM is fully erased." );
  } else {
    Serial.println();
    Serial.println( "This EPROM contains data." );
  }
}

void setAddressForROMTransfer( int address ) {
  // Write 6 bits into the flip-flop.
  if( address & (1<<2) ) digitalWrite(AD2, HIGH); else digitalWrite(AD2, LOW);
  if( address & (1<<3) ) digitalWrite(AD3, HIGH); else digitalWrite(AD3, LOW);
  if( address & (1<<4) ) digitalWrite(AD4, HIGH); else digitalWrite(AD4, LOW);
  if( address & (1<<5) ) digitalWrite(AD5, HIGH); else digitalWrite(AD5, LOW);
  if( address & (1<<6) ) digitalWrite(AD6, HIGH); else digitalWrite(AD6, LOW);
  if( address & (1<<7) ) digitalWrite(AD7, HIGH); else digitalWrite(AD7, LOW);

  // Toggle the clock line to latch in the data.
  delay(1);
  digitalWrite(CLK, HIGH);
  delay(1);
  digitalWrite(CLK, LOW );
  delay(1);

  // Write out the remaining 5 bits.
  if( address & (1<<0) ) digitalWrite(AD0, HIGH); else digitalWrite(AD0, LOW);
  if( address & (1<<1) ) digitalWrite(AD1, HIGH); else digitalWrite(AD1, LOW);
  if( address & (1<<8) ) digitalWrite(AD8, HIGH); else digitalWrite(AD8, LOW);
  if( address & (1<<9) ) digitalWrite(AD9, HIGH); else digitalWrite(AD9, LOW);
  if( address & (1<<10) ) digitalWrite(AD10, HIGH); else digitalWrite(AD10, LOW);
  delay(1);
}

int readByteFromROM() {
  int data = 0;

  // Drop the OE line, since data is available on the falling edge.
  setOEpinLOW();
  delay(1);

  if( digitalRead(DO0) == HIGH ) data |= 1<<0;
  if( digitalRead(DO1) == HIGH ) data |= 1<<1;
  if( digitalRead(DO2) == HIGH ) data |= 1<<2;
  if( digitalRead(DO3) == HIGH ) data |= 1<<3;
  if( digitalRead(DO4) == HIGH ) data |= 1<<4;
  if( digitalRead(DO5) == HIGH ) data |= 1<<5;
  if( digitalRead(DO6) == HIGH ) data |= 1<<6;
  if( digitalRead(DO7) == HIGH ) data |= 1<<7;

  setOEpinHIGH();

  return data;
}

void printPaddedHex( unsigned char data ) {
  if( data < 16 ) Serial.print("0");
  Serial.print(data, HEX);
}

void verifyAllDataFromROM() {
  int address;
  unsigned char databyte, origbyte;
  boolean allDataMatches = true;

  Serial.println();
  Serial.println( "Verifying:" );

  // Read data from the EPROM and compare to data stored in this program.
  // Report discrepancies.
  for( address = 0; address < 2048; address++ ) {
    setAddressForROMTransfer(address);

    databyte = readByteFromROM();
    origbyte = readByteFromProgramData( address );

    if( databyte != origbyte ) {
      allDataMatches = false;

      Serial.print( "Address: " );
      Serial.print( address, HEX );
      Serial.print( " eprom:" );
      Serial.print( databyte, HEX );
      Serial.print( " orig:" );
      Serial.println( origbyte, HEX );
    }
  }

  if( allDataMatches ) {
    Serial.println();
    Serial.println( "Verified." );
  }
}

unsigned char readByteFromProgramData( int address ) {
  char hex[3];

  hex[0] = pgm_read_byte( & settings_data[2*address] );
  hex[1] = pgm_read_byte( & settings_data[2*address+1] );
  hex[2] = '\0';

  return strtol(hex, NULL, 16 );
}

void programAllDataToROM( int retries ) {
  int address;
  unsigned char databyte, origbyte;
  boolean allDataMatches = true;

  Serial.println();
  Serial.println( "Programming:" );

  // Read data from the EPROM and compare to data stored in this program.
  // Report discrepancies.
  for( address = 0; address < 2048; address++ ) {
    // Print address to track progress.
    if( address % 16 == 0 ) {
      Serial.println( address, HEX );
    }

    do {
      retries = retries - 1;

      origbyte = readByteFromProgramData( address );
      databyte = writeByteToROMatAddress( origbyte, address );

      if( databyte == origbyte ) {
	break;
      }
    } while( retries>0 );

    if( databyte != origbyte ) {
      allDataMatches = false;

      Serial.print( "Address: " );
      Serial.print( address, HEX );
      Serial.print( " eprom:" );
      Serial.print( databyte, HEX );
      Serial.print( " orig:" );
      Serial.println( origbyte, HEX );
    }
  }

  if( allDataMatches ) {
    Serial.println();
    Serial.println( "Done." );
  } else {
    Serial.println();
    Serial.println( "Unsuccessful." );
  }
}

int writeByteToROMatAddress( unsigned char byte, int address ) {
  int byteWritten = 0;

  // Set the address pins
  setAddressForROMTransfer(address);
  delay(1);

  // Write out the data byte
  setDataPinsForOutput();
  if( byte & (1<<0)) digitalWrite(DO0, HIGH); else digitalWrite(DO0, LOW);
  if( byte & (1<<1)) digitalWrite(DO1, HIGH); else digitalWrite(DO1, LOW);
  if( byte & (1<<2)) digitalWrite(DO2, HIGH); else digitalWrite(DO2, LOW);
  if( byte & (1<<3)) digitalWrite(DO3, HIGH); else digitalWrite(DO3, LOW);
  if( byte & (1<<4)) digitalWrite(DO4, HIGH); else digitalWrite(DO4, LOW);
  if( byte & (1<<5)) digitalWrite(DO5, HIGH); else digitalWrite(DO5, LOW);
  if( byte & (1<<6)) digitalWrite(DO6, HIGH); else digitalWrite(DO6, LOW);
  if( byte & (1<<7)) digitalWrite(DO7, HIGH); else digitalWrite(DO7, LOW);
  delay(1);

  // Pulse the CE line high for 50ms to write the data.
  digitalWrite(CE, HIGH);
  delay(50);
  digitalWrite(CE, LOW);
  delay(1);

  // Drop the OE line to verify the data
  setDataPinsForInput();
  digitalWrite(_OE, LOW);

  delay(1);

  if( digitalRead(DO0) == HIGH ) byteWritten |= 1<<0;
  if( digitalRead(DO1) == HIGH ) byteWritten |= 1<<1;
  if( digitalRead(DO2) == HIGH ) byteWritten |= 1<<2;
  if( digitalRead(DO3) == HIGH ) byteWritten |= 1<<3;
  if( digitalRead(DO4) == HIGH ) byteWritten |= 1<<4;
  if( digitalRead(DO5) == HIGH ) byteWritten |= 1<<5;
  if( digitalRead(DO6) == HIGH ) byteWritten |= 1<<6;
  if( digitalRead(DO7) == HIGH ) byteWritten |= 1<<7;

  setOEpinHIGH();

  return byteWritten;
}


